""" This module contains functions for loading data. """
from typing import Tuple, Optional, Dict

import arff
import pandas as pd


def arff_to_pandas(file_path: str) -> pd.DataFrame:
    """ Load data from the ARFF file into a pd.DataFrame.

    Parameters
    ----------
    file_path: str
        Path of the ARFF file

    Returns
    -------
    pandas.DataFrame
        A dataframe of the data in the ARFF file,
        with categorical columns having category dtype.
    """
    if not isinstance(file_path, str):
        raise TypeError(f"`file_path` must be of type `str` but is of type {type(file_path)}")

    with open(file_path, 'r') as arff_file:
        arff_dict = arff.load(arff_file)

    attribute_names, data_types = zip(*arff_dict['attributes'])
    data = pd.DataFrame(arff_dict['data'], columns=attribute_names)
    for attribute_name, dtype in arff_dict['attributes']:
        # 'real' and 'numeric' are probably interpreted correctly, date support needs to be added.
        if isinstance(dtype, list):
            data[attribute_name] = data[attribute_name].astype('category')
    return data


def X_y_from_arff(file_path: str, split_column: Optional[str] = None) -> Tuple[pd.DataFrame, pd.Series]:
    """ Load data from the ARFF file into pandas DataFrame and specified column to pd.Series. "

    Parameters
    ----------
    file_path: str
        path to the ARFF file.
    split_column: str, optional (default=None)
        Column to split and return separately (e.g. target column).
        Value should either match a column name or None.
        If None is specified, the last column is returned separately.

    Returns
    -------
    Tuple[pd.DataFrame, pd.Series]
        Features (everything except split_column) and targets (split_column).
    """
    data = arff_to_pandas(file_path)

    if split_column is None:
        return data.iloc[:, :-1], data.iloc[:, -1]
    elif split_column in data.columns:
        return data.loc[:, data.columns != split_column], data.loc[:, split_column]
    else:
        raise ValueError("No column with name {} found in ARFF file {}".format(split_column, file_path))


def load_feature_metadata_from_arff(file_path: str) -> Dict[str, str]:
    """ Load the header of the ARFF file and return the type of each attribute. """
    data_header = '@DATA'
    attribute_indicator = '@ATTRIBUTE'
    attributes = {}
    with open(file_path, 'r') as fh:
        line = fh.readline()
        while not line.startswith(data_header):
            if line.startswith(attribute_indicator):
                # arff uses a space separator, but allows spaces in
                # feature name (name must be quoted) and feature type (if nominal).
                indicator, name_and_type = line.split(' ', 1)
                if name_and_type.startswith('"'):
                    name, data_type = name_and_type[1:].split('" ', 1)
                    name = name
                else:
                    name, data_type = name_and_type.split(' ', 1)
                attributes[name] = data_type
            line = fh.readline()[:-1]  # remove newline character
    return attributes



import os
import time

import dash_table
import dash_core_components as dcc
import dash_html_components as html
import dash_bootstrap_components as dbc
from dash.exceptions import PreventUpdate
from plotly import graph_objects as go
from dash.dependencies import Input, Output, State
import pandas as pd

from gama.dashboard.components.cli_window import CLIWindow
from gama.dashboard.pages.base_page import BasePage
from gama.logging.GamaReport import GamaReport


class RunningPage(BasePage):

    def __init__(self):
        super().__init__(name='Running', alignment=-1)
        self.cli = None
        self.id = 'running-page'
        self.report = None
        self.log_file = None

    def build_page(self, app, controller):
        self.cli = CLIWindow('cli', app)
        plot_area = self.plot_area()
        pl_viz = self.pipeline_viz()
        pl_list = self.pipeline_list()
        ticker = dcc.Interval(id='ticker', interval=5000)
        self._content = html.Div(
            id=self.id,
            children=[
                dbc.Row([
                    dbc.Col(plot_area, width=8),
                    dbc.Col(self.cli.html),
                ]),
                dbc.Row([
                    dbc.Col(pl_viz, width=4),
                    dbc.Col(pl_list)
                ]),
                ticker
            ]
        )

        app.callback(
            [Output('evaluation-graph', 'figure'),
             Output('pipeline-table', 'data'),
             Output('pl-viz', 'children'),
             Output('pipeline-table', 'selected_rows'),
             Output('pipeline-table', 'selected_row_ids'),
             Output('evaluation-graph', 'clickData')],
            [Input('ticker', 'n_intervals'),
             Input('running-page-store', 'data')]
        )(self.update_page)

        app.callback(
            [Output('running-page-store', 'data')],
            [Input('evaluation-graph', 'clickData'),
             Input('pipeline-table', 'selected_row_ids')],
            [State('running-page-store', 'data')]
        )(self.update_selection)

        return self._content

    def update_selection(self, click_data, selected_row_ids, page_store):
        cell_selected = None if selected_row_ids is None else selected_row_ids[0]
        click_selected = None if click_data is None else click_data["points"][0]['customdata']
        # Selected row ids and click data are always set back to None. The value that is not None is the new value.
        if click_data is not None or cell_selected is not None:
            self.need_update = True
            page_store['selected_pipeline'] = click_selected if click_selected is not None else cell_selected
            return [page_store]
        # First call or sync call.
        raise PreventUpdate

    def update_page(self, _, page_store):
        start_update = time.time()
        selected_pipeline = page_store.get('selected_pipeline', None)
        if ((self.report is None and (self.log_file is None or not os.path.exists(self.log_file)))
                or (self.report is not None and not self.report.update() and not self.need_update)):
            # The report does not exist, or exists but nothing is updated.
            raise PreventUpdate
        elif self.report is None:
            self.report = GamaReport(self.log_file)

        with pd.option_context('mode.use_inf_as_na', True):
            evaluations = self.report.evaluations.dropna()

        self.need_update = False
        scatters = self.scatter_plot(evaluations, self.report.metrics, selected_pipeline)
        metric_one, metric_two = self.report.metrics
        metric_one_text = metric_one.replace('_', ' ')

        figure = {
            'data': scatters,
            'layout': dict(
                hovermode='closest',
                clickmode='event+select',
                title=f'{metric_one_text} vs {metric_two}',
                xaxis=dict(title=metric_one_text),
                yaxis=dict(title=metric_two, tickformat=',d'),  # tickformat forces integer ticks for length,
                uirevision='never_reset_zoom'
            )
        }

        pl_table_data = [{'pl': self.report.individuals[id_].short_name(' > '),
                          'id': id_,
                          'score': score}
                         for id_, score in zip(evaluations.id, evaluations[metric_one])]
        row_id = [i for i, id_ in enumerate(evaluations.id) if id_ == selected_pipeline]

        def format_pipeline(ind):
            pipeline_elements = []
            for primitive_node in reversed(ind.primitives):
                pipeline_elements.append(html.B(str(primitive_node._primitive)))
                pipeline_elements.append(html.Br())
                for terminal in primitive_node._terminals:
                    pipeline_elements.append(f'    {terminal}')
                    pipeline_elements.append(html.Br())
            return pipeline_elements
        pl_viz_data = None if selected_pipeline is None else format_pipeline(self.report.individuals[selected_pipeline])

        print('Update complete in ', time.time() - start_update)
        return figure, pl_table_data, pl_viz_data, row_id, None, None

    def scatter_plot(self, evaluations, metrics, selected_pipeline: str = None):
        metric_one, metric_two = metrics

        # Marker size indicates recency of the evaluations, recent evaluations are bigger.
        biggest_size = 25
        smallest_size = 5
        selected_size = 30
        d_size_min_max = biggest_size - smallest_size

        sizes = list(range(smallest_size, biggest_size))[-len(evaluations):]
        if len(evaluations) > d_size_min_max:
            sizes = [smallest_size] * (len(evaluations) - d_size_min_max) + sizes
        if selected_pipeline is not None:
            sizes = [size if id_ != selected_pipeline else selected_size
                     for size, id_ in zip(sizes, evaluations.id)]

        default_color = '#301cc9'
        selected_color = '#c81818'

        colors = [default_color if id_ != selected_pipeline else selected_color
                  for id_ in evaluations.id]

        all_scatter = go.Scatter(
            x=evaluations[metric_one],
            y=-evaluations[metric_two],
            mode='markers',
            marker={'color': colors, 'size': sizes},
            name='all evaluations',
            text=[self.report.individuals[id_].short_name() for id_ in evaluations.id],
            customdata=evaluations.id,
        )
        return [all_scatter]

    def gama_started(self, process, log_file):
        self.cli.monitor(process)
        self.log_file = log_file

    def plot_area(self):
        scatter = dcc.Graph(
            id='evaluation-graph',
            figure={
                'data': [],
                'layout': dict(
                    hovermode='closest',
                    transition={'duration': 500},
                )
            }
        )
        return html.Div(scatter, style={'height': '100%', 'box-shadow': '1px 1px 1px black', 'padding': '2%'})

    def pipeline_list(self):
        ta = dash_table.DataTable(
            id='pipeline-table',
            columns=[{'name': 'Pipeline', 'id': 'pl'}, {'name': 'Score', 'id': 'score'}],
            data=[],
            style_table={
                'maxHeight': '300px',
                'overflowY': 'scroll'
            },
            filter_action="native",
            sort_action="native",
            row_selectable='single',
            persistence_type='session',
            persistence=True
        )

        return html.Div(ta, style={'height': '100%', 'box-shadow': '1px 1px 1px black', 'padding': '2%'})

    def pipeline_viz(self):
        return html.Div(id='pl-viz', style={'height': '100%', 'box-shadow': '1px 1px 1px black', 'padding': '2%', 'whiteSpace': 'pre-wrap'})



""" Implementation of NSGA-II and its subroutines as defined in Deb et al. (2002)

Deb, K., Pratap, A., Agarwal, S., & Meyarivan, T. A. M. T. (2002).
A fast and elitist multiobjective genetic algorithm: NSGA-II.
IEEE transactions on evolutionary computation, 6(2), 182-197.
"""
import itertools
import random
from functools import cmp_to_key
from typing import List, Any, Callable
import numpy as np


class NSGAMeta:

    def __init__(self, obj, metrics):
        self.obj = obj
        self.values = tuple((m(obj) for m in metrics))
        self.rank = None
        self.distance = 0
        self.dominating = []
        self.domination_counter = 0

    def dominates(self, other: 'NSGAMeta'):
        for self_val, other_val in zip(self.values, other.values):
            if self_val <= other_val:  # or maybe <?
                return False
        return True

    def crowd_compare(self, other: 'NSGAMeta'):
        """ Favor higher rank, if equal, favor less crowded. """
        self_better = (self.rank < other.rank or
                       (self.rank == other.rank and self.distance > other.distance))
        return -1 if self_better else 1


def nsga2_select(population: List[Any], n: int, metrics: List[Callable[[Any], float]]) -> List[Any]:
    """ Select n pairs from the population through binary tournament selection based on crowding distance. """
    # Entire population is returned, but with rank and distance information.
    candidates = nsga2(population, n=len(population), metrics=metrics, return_meta=True)

    def select_one():
        ind1, ind2 = random.sample(candidates, k=2)
        return ind1 if ind1.crowd_compare(ind2) < 0 else ind2

    selected = []
    for _ in range(n):
        selected.append((select_one().obj, select_one().obj))
    return selected


def nsga2(population: List[Any], n: int, metrics: List[Callable[[Any], float]], return_meta: bool=False) -> List[Any]:
    """ Selects n individuals from the population to create offspring with according to NSGA-II.

    Parameters
    ----------
    population: List[T]
        A list of objects.
    n: int
        Number of objects to pick out of population. Must be greater than 0 and smaller than len(population).
    metrics: List[Callable[[T], float]]
        List of functions which obtain the values for each dimension on which to compare elements of population.
    return_meta: bool (default=False)
        If True, return the selected individuals wrapped in a NSGAMeta class with information such as rank and distance.
        If False, return the selected individuals as they were passed to this function.

    Returns
    -------
    selection: List[T]
        A list of size n containing a subset of population.
    """
    if n == 0 or n > len(population):
        raise ValueError(f"{n} is not a valid value for `n`, must be 0 < n < len(population) ({len(population)}).")
    population = [NSGAMeta(p, metrics) for p in population]
    selection = []
    fronts = fast_non_dominated_sort(population)
    i = 0

    while len(selection) < n:
        crowding_distance_assignment(fronts[i])
        if len(selection) + len(fronts[i]) < n:
            # Enough space for the entire Pareto front, include all
            selection += fronts[i]
        else:
            # Only the least crowded remainder is selected
            s = sorted(fronts[i], key=cmp_to_key(lambda x, y: x.crowd_compare(y)))
            selection += s[: (n - len(selection))]  # Fill up to n
        i += 1

    return selection if return_meta else [s.obj for s in selection]


def fast_non_dominated_sort(P: List[NSGAMeta]) -> List[List[NSGAMeta]]:
    """ Sorts P into Pareto fronts. """
    fronts = [[]]
    for p, q in itertools.combinations(P, 2):
        if p.dominates(q):
            p.dominating.append(q)
            q.domination_counter += 1
        elif q.dominates(p):
            q.dominating.append(p)
            p.domination_counter += 1

    for p in P:
        if p.domination_counter == 0:
            p.rank = 1
            fronts[0].append(p)

    i = 0
    while fronts[i] != []:
        fronts.append([])
        for p in fronts[i]:
            for q in p.dominating:
                q.domination_counter -= 1
                if q.domination_counter == 0:
                    q.rank = i + 1
                    fronts[i + 1].append(q)
        i += 1
    return fronts


def crowding_distance_assignment(I: List[NSGAMeta]) -> None:
    for m in range(len(I[0].values)):
        I = sorted(I, key=lambda x: x.values[m])
        I[0].distance = I[-1].distance = float('inf')
        if I[-1].values[m] == I[0].values[m] or np.isinf(I[0].values[m]) or np.isinf(I[-1].values[m]):
            # Would raise divisionbyzero later, or give other numerical warnings.
            # This typically happens only for the worst pareto front(s), so the inaccuracy in crowding distance for
            # the remainder is not really a concern. Might consider immediately removing failing individuals.
            continue

        for i_prev, i, i_next in zip(I, I[1:], I[2:]):
            i.distance += (i_next.values[m] - i_prev.values[m]) / (I[-1].values[m] - I[0].values[m])



from typing import Tuple, List

from sklearn.pipeline import Pipeline

from gama.genetic_programming.components import Individual


def model_to_python(pipeline: Pipeline) -> str:
    """ Generates Python code which sets up the machine learning pipeline represented by `individual`. """
    imports = []
    for name, step in pipeline.steps:
        imports.append(f"from {step.__module__} import {step.__class__.__name__}")

    script = ("from sklearn.pipeline import Pipeline\n" +
              '\n'.join(imports) + '\n\n' +
              'pipeline = ' + str(pipeline) + '\n')

    return script


def imports_and_steps_for_individual(individual: Individual) -> Tuple[List[str], List[Tuple[str, str]]]:
    """ Returns a list of import statements, and a list of tuples with pipeline step names and step values.

    E.g. (["from sklearn.naive_bayes import GaussianNB"], [('0', 'GaussianNB()')])
    """
    imports = ["from numpy import nan", "from sklearn.pipeline import Pipeline"]
    for name, step in individual.pipeline.steps:
        imports.append(f"from {step.__module__} import {step.__class__.__name__}")

    # The pipeline consists of two steps:
    # - Data Preparation: SimpleImputer and possibly One-hot or Target encoding.
    # - The remainder: additional transformers and an estimator.
    # Because the data preparation step is (currently) defined once and shared across all pipelines,
    # this is not captured in the Individual itself, so we have to extract it from the compiled pipeline.
    steps = []
    n_data_preparation_steps = len(individual.pipeline.steps) - len(individual.primitives)
    for name, step in individual.pipeline.steps[:n_data_preparation_steps]:
        steps.append((name, step))

    for i, primitive_node in reversed(list(enumerate(individual.primitives))):
        steps.append((str(i), primitive_node.str_nonrecursive))
        for terminal in primitive_node._terminals:
            if callable(terminal.value):
                imports.append(f"from {terminal.value.__module__} import {terminal.value.__name__}")

    return imports, steps


def individual_to_python(individual: Individual) -> str:
    """ Generates Python code which sets up the machine learning pipeline represented by `individual`. """
    imports, steps = imports_and_steps_for_individual(individual)
    steps_str = ',\n'.join([f"('{name}', {step})" for name, step in steps])
    pipeline = f"Pipeline([{steps_str}])"
    script = ("from sklearn.pipeline import Pipeline\n" +
              '\n'.join(imports) + '\n\n' +
              'pipeline = ' + pipeline + '\n')

    return script



""" Provides updates on GAMA's search. Next step is to call GAMA directly, but the `fit` call has to be made async. """

import shlex
import subprocess
from collections import defaultdict


class Controller:

    def __init__(self):
        self._subscribers = defaultdict(list)

    def start_gama(
            self, n_clicks, metric, regularize, n_jobs,
            max_total_time_h, max_total_time_m, max_eval_time_h, max_eval_time_m,
            input_file, log_file):
        # For some reason, 0 input registers as None.
        max_total_time_h = 0 if max_total_time_h is None else max_total_time_h
        max_total_time_m = 0 if max_total_time_m is None else max_total_time_m
        max_eval_time_h = 0 if max_eval_time_h is None else max_eval_time_h
        max_eval_time_m = 0 if max_eval_time_m is None else max_eval_time_m
        max_total_time = (max_total_time_h * 60 + max_total_time_m)
        max_eval_time = (max_eval_time_h * 60 + max_eval_time_m)
        command = f'gama "{input_file}" -v -n {n_jobs} -t {max_total_time} --time_pipeline {max_eval_time} -log {log_file}'
        if regularize != 'on':
            command += ' --long'
        if metric != 'default':
            command += f' -m {metric}'

        command = shlex.split(command)
        # fake_command = ['python', '-h']
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self._on_gama_started(process, log_file)

    def _on_gama_started(self, process, log_file):
        for subscriber in self._subscribers['gama_started']:
            subscriber(process, log_file)

    def gama_started(self, callback_function):
        self._subscribers['gama_started'].append(callback_function)

    def gama_ended(self, callback_function):
        self._subscribers['gama_ended'].append(callback_function)



from typing import List

try:
    import dash
    import dash_core_components as dcc
    import dash_bootstrap_components as dbc
    import dash_html_components as html
    from dash.dependencies import Input, Output
except ImportError:
    print("Unable to import Dash modules. Make sure GAMA is installed with the optional visualization dependencies."
          "E.g.: pip install gama[vis]")
    raise

from gama.dashboard.controller import Controller
from gama.dashboard.pages.base_page import BasePage


external_scripts = ['https://code.jquery.com/jquery-3.3.1.min.js',
                    'https://cdn.datatables.net/v/dt/dt-1.10.18/datatables.min.js']
dashboard = dash.Dash('GamaDashboard', external_stylesheets=[dbc.themes.BOOTSTRAP], external_scripts=external_scripts)
dashboard.config.suppress_callback_exceptions = True

controller = Controller()


# === Construct UI elements ===

def build_app():
    from gama.dashboard.pages import pages
    base = create_generic_layout()
    base['tabs'].children = create_tabs(pages)
    page_stores = []
    for page in pages:
        page.build_page(dashboard, controller)
        if hasattr(page, 'gama_started'):
            controller.gama_started(page.gama_started)
        page_stores.append(dcc.Store(id=f'{page.id}-store', storage_type='session', data={}))
    base['page-stores'].children = page_stores
    return base


def create_generic_layout():
    """ Creates the generic layout of tabs and their content pages. """
    tab_banner_style = {
        'border-top-left-radius': '3px',
        'background-color': '#f9f9f9',
        'padding': '0px 24px',
        'border-bottom': '1px solid #d6d6d6'
    }

    return html.Div(
        id="page",
        children=[
            html.Div(id="tabs", style=tab_banner_style),
            html.Div(id="content"),
            html.Div(id="page-stores")
        ],
        style={'font-family': "'Open Sans Semi Bold', sans-serif"}
    )


def create_tabs(pages: List[BasePage]):
    if pages == []:
        raise ValueError("Must have at least one tab.")

    # Sort pages by alignment
    sorted_pages = sorted(pages, key=lambda p: p.alignment)
    grouped_pages = ([page for page in sorted_pages if page.alignment >= 0] +
                     [page for page in sorted_pages if page.alignment < 0])
    tabs = [create_tab(page.name) for page in grouped_pages]
    return [dcc.Tabs(id="page-tabs", value=tabs[0].value, children=tabs)]


def create_tab(name: str):
    tab_style = {
        'color': 'black',
        'width': '10%',
        'border-top-left-radius': '3px',  # round tab corners
        'border-top-right-radius': '3px',
        'border-bottom': '0px',  # bottom box-shadow still present
        'padding': '6px'
    }
    selected_tab_style = {
        ** tab_style,
        'border-top': '3px solid #c81818',  # Highlight color (TU/e colored)
        'box-shadow': '1px 1px 0px white'  # removes bottom edge
    }
    return dcc.Tab(
        id=f"{name}-tab",
        label=name,
        value=name,
        style=tab_style,
        selected_style=selected_tab_style
    )


# === Callbacks ===

@dashboard.callback(
    [Output("content", "children")],
    [Input("page-tabs", "value")]
)
def display_page_content(page_name):
    from gama.dashboard.pages import pages
    page = [page for page in pages if page.name == page_name][0]
    page.need_update = True
    return [page.content]


def main(open_window: bool = True, **server_kwargs):
    dashboard.layout = build_app()
    if open_window:
        import webbrowser
        from threading import Timer
        port = server_kwargs.get('port', 8050)
        Timer(1, webbrowser.open_new, kwargs=dict(url=f"http://127.0.0.1:{port}")).start()
    dashboard.run_server(**server_kwargs)


if __name__ == '__main__':
    main(debug=True)



from typing import Dict

from dash.dependencies import Input, Output, State
import dash_core_components as dcc
import dash_bootstrap_components as dbc


def automark_slider(app: 'Dash', id_: str, label: str, slider_kwargs: Dict):
    defaults = dict(min=1, max=10, value=1, updatemode='drag')
    defaults.update(slider_kwargs)
    marks = {defaults['min']: defaults['min'], defaults['max']: defaults['max']}

    cpu_input = dbc.FormGroup(
        [
            dbc.Label(label, html_for=id_, width=6),
            dbc.Col(
                dcc.Slider(id=id_, marks=marks, **defaults)
            )
        ],
        row=True
    )
    app.callback.append(
        Output(id_, "marks"),
        [Input(id_, "value")],
        [State(id_, "min"), State(id_, "max")]
    )(_update_marks)
    return cpu_input


def _update_marks(selected_value, min_, max_):
    return {min_: str(min_), selected_value: str(selected_value), max_: str(max_)}


class ToggleButton:

    def __init__(self, button_id: str, app: 'Dash', label: str, start_on: bool = True):
        self._button_id = button_id
        self.html = self._build_content(label, start_on)

    def _build_content(self, label: str, start_on: bool):
        return dbc.FormGroup(
            [
                dbc.Label(label, html_for=self._button_id, width=6),
                dbc.Col(dbc.Checklist(id=self._button_id, options=[{'label': '', 'value': 'on'}], switch=True,
                                      value='on' if start_on else 'off')),
            ],
            row=True
        )



import abc


class BasePage(abc.ABC):

    def __init__(self, name: str, alignment: int):
        """ Defines the basic behavior of a page.

        Parameters
        ----------
        name: str
            Name of the page, displayed in the tab.

        alignment: int
            Defines the order of tabs.
            Positive numbers are aligned to the left, negative to the right.
            Within the groups, bigger numbers are placed to the right.
            E.g.: [0][1][2] ... [-2][-1]
        """
        self.name = name
        self.alignment = alignment
        self.need_update = False
        self._content = None

    @property
    def content(self):
        return self._content

    @abc.abstractmethod
    def build_page(self, app, controller):
        """ Populate the `content` field with html, register any callbacks. """
        raise NotImplementedError



import inspect
from typing import Union, Optional

import numpy as np
import pandas as pd
from sklearn.base import ClassifierMixin
from sklearn.preprocessing import LabelEncoder

from .gama import Gama
from gama.data import X_y_from_arff
from gama.configuration.classification import clf_config
from gama.utilities.metrics import scoring_to_metric


class GamaClassifier(Gama):
    """ Wrapper for the toolbox logic executing the AutoML pipeline for (multi-class) classification. """
    def __init__(self, config=None, scoring='neg_log_loss', *args, **kwargs):
        if not config:
            # Do this to avoid the whole dictionary being included in the documentation.
            config = clf_config

        self._metrics = scoring_to_metric(scoring)
        if any(metric.requires_probabilities for metric in self._metrics):
            # we don't want classifiers that do not have `predict_proba`, because then we have to
            # start doing one hot encodings of predictions etc.
            config = {alg: hp for (alg, hp) in config.items()
                      if not (inspect.isclass(alg) and issubclass(alg, ClassifierMixin)and not hasattr(alg(), 'predict_proba'))}

        self._label_encoder = None
        super().__init__(*args, **kwargs, config=config, scoring=scoring)

    def _predict(self, x: pd.DataFrame):
        """ Predict the target for input X.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            array with predictions of shape (N,) where N is the length of the first dimension of X.
        """
        y = self.model.predict(x)
        # Decode the predicted labels - necessary only if ensemble is not used.
        if y[0] not in self._label_encoder.classes_:
            y = self._label_encoder.inverse_transform(y)
        return y

    def _predict_proba(self, x: pd.DataFrame):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame
            A dataframe with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is the length of the
            first dimension of x, and K is the number of class labels found in `y` of `fit`.
        """
        return self.model.predict_proba(x)

    def predict_proba(self, x: Union[pd.DataFrame, np.ndarray]):
        """ Predict the class probabilities for input x.

        Predict target for x, using the best found pipeline(s) during the `fit` call.

        Parameters
        ----------
        x: pandas.DataFrame or numpy.ndarray
            A dataframe or numpy array with the same number of columns as the input to `fit`.

        Returns
        -------
        numpy.ndarray
            Array of shape (N, K) with class probabilities where N is the length of the
            first dimension of x, and K is the number of class labels found in `y` of `fit`.
        """
        if isinstance(x, np.ndarray):
            x = pd.DataFrame(x)
            for col in self._X.columns:
                x[col] = x[col].astype(self._X[col].dtype)
        return self._predict_proba(x)

    def predict_proba_arff(self, arff_file_path: str, target_column: Optional[str] = None):
        """ Predict the class probabilities for input in the arff_file, must have empty target column.

        Parameters
        ----------
        arff_file_path: str
            An ARFF file with the same columns as the one that used in fit.
            Target column must be present in file, but its values are ignored (can be '?').
        target_column: str, optional (default=None)
            Specifies which column the model should predict.
            If left None, the last column is taken to be the target.

        Returns
        -------
        numpy.ndarray
            Numpy array with class probabilities. The array is of shape (N, K) where N is the length of the
            first dimension of X, and K is the number of class labels found in `y` of `fit`.
        """
        X, _ = X_y_from_arff(arff_file_path, target_column)
        return self._predict_proba(X)

    def fit(self, x, y, *args, **kwargs):
        """ Should use base class documentation. """
        y_ = y.squeeze() if isinstance(y, pd.DataFrame) else y
        self._label_encoder = LabelEncoder().fit(y_)
        if any([isinstance(yi, str) for yi in y_]):
            # If target values are `str` we encode them or scikit-learn will complain.
            y = self._label_encoder.transform(y_)
        super().fit(x, y, *args, **kwargs)

    def _encode_labels(self, y):
        self._label_encoder = LabelEncoder().fit(y)
        return self._label_encoder.transform(y)



import argparse
import logging
import pickle
import pandas as pd

from gama import GamaClassifier, GamaRegressor
from gama.data import load_feature_metadata_from_arff


def parse_args():
    parser = argparse.ArgumentParser(
        description="An AutoML tool that optimizes machine learning pipelines for your data."
    )

    parser.add_argument(
        'input_file',
        type=str,
        help='An ARFF file with the data to optimize a model for.'
    )

    io_group = parser.add_argument_group("File I/O")

    # io_group.add_argument(
    #     '-sep',
    #     dest='separator',
    #     type=str,
    #     default=',',
    #     help="For CSV files only, the character used to separate columns in the input file. (default=',')"
    # )
    io_group.add_argument(
        '--target',
        dest='target',
        type=str,
        default=None,
        help="The target column in the input file."
    )
    io_group.add_argument(
        '-o',
        dest='output_file',
        type=str,
        default='gama_model.pkl',
        help="Path to store the final model in. (default=gama_model.pkl)"
    )

    io_group.add_argument(
        '-py',
        dest='export_python',
        type=str,
        default=None,
        help="If set, will export a Python file to this destination which sets up the final model."
    )

    optimization = parser.add_argument_group("Optimization")
    optimization.add_argument(
        '--mode',
        dest='mode',
        type=str,
        default=None,
        help="By default, it is determined automatically from the input file."
             "Use this parameter force 'classification' or 'regression' mode. "
    )
    optimization.add_argument(
        '-m',
        dest='metric',
        type=str,
        default=None,
        help="The metric to optimize the model for. "
             "Default is log loss for classification and RMSE for regression."
    )
    optimization.add_argument(
        '--long',
        dest='prefer_short',
        action='store_false',
        help="By default GAMA will guide search towards shorter pipelines. "
             "Set this flag to disable the feature."
    )

    resources = parser.add_argument_group("Resources")
    resources.add_argument(
        '-t',
        dest='time_limit_m',
        type=int,
        default=60,
        help="The maximum time (in minutes) GAMA has for optimization. (default=60)"
    )
    resources.add_argument(
        '--time_pipeline',
        dest='max_eval_time_m',
        type=int,
        default=5,
        help="The maximum time (in minutes) GAMA may spend one a single pipeline evaluation. (default=5)"
    )
    resources.add_argument(
        '-n',
        dest='n_jobs',
        type=int,
        default=1,
        help="The number of jobs to run in parallel for pipeline evaluations. (default=1)"
    )

    # Extra
    parser.add_argument(
        '-log',
        dest='logpath',
        default=None,
        type=str,
        help="File to store GAMA's log."
    )
    parser.add_argument(
        '-v',
        dest='verbose',
        action='store_true',
        help="Report status updates to console."
    )

    return parser.parse_args()


def main():
    args = parse_args()

    print('CLI: Processing input')
    if args.input_file.lower().endswith('.csv'):
        raise NotImplementedError("CSV currently not supported.")
        # data = pd.read_csv(args.input_file, sep=args.separator)
    if args.input_file.lower().endswith('.arff') and args.mode is None:
        attributes = load_feature_metadata_from_arff(args.input_file)
        target = list(attributes)[-1] if args.target is None else args.target
        target_type = attributes[target]
        if '{' in target_type:
            # Nominal features are denoted by listen all their values, eg. {VALUE_1, VALUE_2, ...}
            args.mode = 'classification'
        elif target_type.lower() == 'real':
            args.mode = 'regression'
        else:
            raise ValueError(f"Target column {target} has type {target_type}, which GAMA can't model.")

    print('CLI: Initializing GAMA')
    log_level = logging.INFO if args.verbose else logging.WARNING
    configuration = dict(
        regularize_length=args.prefer_short,
        max_total_time=args.time_limit_m * 60,
        max_eval_time=args.max_eval_time_m * 60,
        n_jobs=args.n_jobs,
        verbosity=log_level,
        keep_analysis_log=args.logpath
    )
    if args.metric:
        configuration['scoring'] = args.metric

    if args.mode == 'regression':
        automl = GamaRegressor(**configuration)
    elif args.mode == 'classification':
        automl = GamaClassifier(**configuration)
    else:
        raise ValueError(f"Mode {args.mode} is not valid (--mode).")

    print('CLI: Starting model search')
    if args.input_file.lower().endswith('.arff'):
        automl.fit_arff(args.input_file.lower(), target_column=args.target)
    #else:
    #    automl.fit(x, y)

    # == Model Export ===
    print('CLI: Exporting models.')
    with open(args.output_file, 'wb') as fh:
        pickle.dump(automl.model, fh)

    if args.export_python is not None:
        automl.export_script(args.export_python, raise_if_exists=False)
    print('done!')


if __name__ == '__main__':
    main()



import multiprocessing
import os
from typing import Optional, List, Dict

import dash_table
import dash_core_components as dcc
import dash_bootstrap_components as dbc
import dash_html_components as html
from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate

from gama.dashboard.pages.base_page import BasePage


class HomePage(BasePage):
    callbacks = []

    def __init__(self):
        super().__init__(name='Home', alignment=0)
        self.id = "home-page"

    def build_page(self, app, controller):
        self._build_content(app, controller)
        if app is not None:
            self._register_callbacks(app)

    def _build_content(self, app, controller) -> html.Div:
        """ Build all the components of the page. """
        configuration = build_configuration_menu(app, controller)
        configuration.style['width'] = '35%'
        configuration.style['float'] = 'left'
        data_navigator = build_data_navigator()
        data_navigator.style['width'] = '65%'
        data_navigator.style['float'] = 'right'
        self._content = html.Div(
            id=self.id,
            children=[
                configuration,
                data_navigator
            ]
        )
        return self._content

    def _register_callbacks(self, app):
        for (io, fn) in HomePage.callbacks:
            app.callback(*io)(fn)
        HomePage.callbacks = []

    def load(self):
        pass

    def unload(self):
        pass

# === Configuration Menu ===

def cpu_slider():
    n_cpus = multiprocessing.cpu_count()
    id_ = 'cpu_slider'
    cpu_input = dbc.FormGroup(
        [
            dbc.Label("N Jobs", html_for=id_, width=6),
            dbc.Col(
                dcc.Slider(id=id_, min=1, max=n_cpus, updatemode='drag',
                           value=1, marks={1: '1', n_cpus: str(n_cpus)},
                           persistence_type='session', persistence=True)
            )
        ],
        row=True
    )
    HomePage.callbacks.append(((
        Output(id_, "marks"),
        [Input(id_, "value")],
        [State(id_, "min"), State(id_, "max")]),
        update_marks
    ))
    return cpu_input


def time_nud(label_text: str, hour_id: str, hour_default: int, minute_id: str, minute_default: int):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=hour_id, width=6),
            dbc.Col(
                dbc.InputGroup(
                    [
                        dbc.Input(id=hour_id, type='number', min=0, max=99, step=1, value=hour_default),
                        dbc.InputGroupAddon("H", addon_type="append"),
                    ]
                )
            ),
            dbc.Col(
                dbc.InputGroup(
                    [
                        dbc.Input(id=minute_id, type='number', min=0, max=59, step=1, value=minute_default),
                        dbc.InputGroupAddon("M", addon_type="append"),
                    ]
                )
            )
        ],
        row=True
    )


def toggle_button(label_text: str, id_: str, start_on: bool = True):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(dbc.Checklist(id=id_, options=[{'label': '', 'value': 'on'}], switch=True,
                                  value='on' if start_on else 'off')),
        ],
        row=True
    )


def text_input(label_text: str, default_text: str, id_: str):
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(dbc.Input(id=id_, type='text', placeholder=default_text, value=default_text)),
        ],
        row=True
    )


def dropdown(label_text: str, id_: str, options: Dict[str, str], value: Optional[str] = None):
    """ options formatted as {LABEL_KEY: LABEL_TEXT, ...} """
    return dbc.FormGroup(
        [
            dbc.Label(label_text, html_for=id_, width=6),
            dbc.Col(
                dcc.Dropdown(
                    id=id_,
                    options=[
                        {'label': text, 'value': key}
                        for key, text in options.items()
                    ],
                    clearable=False,
                    value=value,
                    persistence_type='session',
                    persistence=True
                ),
            ),
        ],
        row=True
    )


def button_header(text: str, id_: str, level: int = 4):
    header = f"{'#' * level} {text}"
    return dbc.FormGroup([dbc.Button([dcc.Markdown(header)], id=id_, block=True, color='primary')])


def markdown_header(text: str, level: int = 4, with_horizontal_rule: bool = True):
    header = f"{'#' * level} {text}"
    hr = f"\n{'-'*(level + 1 + len(text))}"  # matching length '-' not required but nicer.
    return dcc.Markdown(f"{header}{hr if with_horizontal_rule else ''}")


def toggle_collapse(click, is_open: bool):
    if click:
        return not is_open
    return is_open


def collapsable_section(header: str, controls: List[dbc.FormGroup], start_open: bool = True):
    header_id = f"{header}-header"
    form_id = f"{header}-form"

    form_header = button_header(header, id_=header_id)
    collapsable_form = dbc.Collapse(
        id=form_id,
        children=[dbc.Form(controls)],
        is_open=start_open
    )

    HomePage.callbacks.append(((
        Output(form_id, "is_open"),
        [Input(header_id, "n_clicks")],
        [State(form_id, "is_open")]),
        toggle_collapse
    ))
    return form_header, collapsable_form


def build_configuration_menu(app, controller) -> html.Div:
    # Optimization
    from gama.utilities.metrics import all_metrics
    metrics = {m: m.replace('_', ' ') for m in all_metrics}
    metrics.update({'default': 'default'})
    scoring_input = dropdown('Metric', 'metric_dropdown', options=metrics, value='default')
    regularize_input = toggle_button('Prefer short pipelines', 'regularize_length_switch')
    optimization = collapsable_section("Optimization", [scoring_input, regularize_input])

    # Resources
    cpu_input = cpu_slider()
    max_total_time_input = time_nud('Max Runtime',
                                    hour_id='max_total_h',
                                    hour_default=1,
                                    minute_id='max_total_m',
                                    minute_default=0)
    max_eval_time_input = time_nud('Max time per pipeline',
                                   hour_id='max_eval_h',
                                   hour_default=0,
                                   minute_id='max_eval_m',
                                   minute_default=5)
    resources = collapsable_section("Resources", [cpu_input, max_total_time_input, max_eval_time_input])

    # Advanced
    log_path = text_input("Logfile Path", 'gama.log', 'logpath')
    advanced = collapsable_section("Advanced", [log_path], start_open=False)

    # Go!
    go_button = dbc.Button([dcc.Markdown("#### Go!")], id='go-button', block=True, color='success', disabled=True)

    def start_gama(*args, **kwargs):
        controller.start_gama(*args, **kwargs)
        return 'danger', dcc.Markdown("#### Stop!"), "Running"

    app.callback(
        [Output('go-button', "color"),
         Output('go-button', "children"),
         Output("page-tabs", "value")],
        [Input('go-button', "n_clicks")],
        [State('metric_dropdown', "value"),
         State('regularize_length_switch', "value"),
         State('cpu_slider', "value"),
         State('max_total_h', "value"),
         State('max_total_m', "value"),
         State('max_eval_h', "value"),
         State('max_eval_m', "value"),
         State('file-path-input', 'value'),
         State('logpath', 'value')]
    )(start_gama)

    return html.Div(
        children=[markdown_header('Configure GAMA', level=2),
                  *optimization,
                  *resources,
                  *advanced,
                  go_button],
        style={'box-shadow': '1px 1px 1px black', 'padding': '2%'}
    )


def update_marks(selected_value, min_, max_):
    return {min_: str(min_), selected_value: str(selected_value), max_: str(max_)}


def build_data_navigator() -> html.Div:
    # Unfortunately, at this time we need a full path, which is not available with dcc.Upload
    #
    # default_message = ['Drag and Drop or ', html.A('Select Files')]
    # upload_file = dcc.Upload(
    #     id='upload-data',
    #     children=html.Div(
    #         id='upload-data-text',
    #         children=default_message
    #     ),
    #     style={'borderWidth': '1px', 'borderRadius': '5px', 'borderStyle': 'dashed', 'textAlign': 'center'}
    # )

    # def update_output(list_of_contents, filename, list_of_dates):
    #     if list_of_contents is not None:
    #         children = html.Div([f"Showing '{filename}'", html.A(' (Click to change)')])
    #         return children, filename
    #     return default_message, filename
    #
    # ---------------------------------------------

    upload_file = dbc.Input(
        id='file-path-input',
        placeholder='Path to data file, e.g. ~/data/mydata.arff',
        type='text'
    )

    table = dash_table.DataTable(
        id='data-table'
    )

    table_container = html.Div(
        id='table-container',
        children=['No data loaded.']
    )

    def update_data_table(filename):
        if filename is not None and os.path.isfile(filename):
            return 'found a file!', False
        return filename, True

    HomePage.callbacks.append(((
        [Output('table-container', 'children'),
         Output('go-button', 'disabled')],
        [Input('file-path-input', 'value')]),
        update_data_table
    ))

    return html.Div(
        children=[markdown_header("Data Navigator", level=2),
                  upload_file,
                  table_container
                  ],
        style={'box-shadow': '1px 1px 1px black', 'padding': '2%'}
    )



import shlex
import subprocess
import threading
import queue

from dash.dependencies import Input, Output, State
from dash.exceptions import PreventUpdate
import dash_core_components as dcc
import dash_html_components as html
import visdcc


def enqueue_output(out, queue_: queue.Queue):
    for line in iter(out.readline, b''):
        queue_.put(line)
    out.close()


class CLIWindow:
    """ A Component for Dash App which simulates a console window.


    """
    def __init__(self, id_: str, app: 'Dash', update_interval_s: float = 1.0, auto_scroll: bool = True):
        self._update_interval_s = update_interval_s
        self.auto_scroll = auto_scroll

        self.console_id = f'{id_}-text'
        self.timer_id = f'{id_}-interval'
        self.js_id = f'{id_}-js'
        self.id_ = id_

        self.autoscroll_script = f"document.getElementById('{self.console_id}').scrollTop = document.getElementById('{self.console_id}').scrollHeight"
        self.process = None
        self._thread = None
        self._queue = None
        self._lines = []

        self.html = self._build_component()
        self._register_callbacks(app)

    def _build_component(self) -> html.Div:
        timer = dcc.Interval(
            id=self.timer_id,
            interval=self._update_interval_s * 1000,
            n_intervals=0
        )
        scroller = visdcc.Run_js(id=self.js_id, run='')
        self.console = dcc.Textarea(
            id=self.console_id,
            contentEditable='false',
            style={'height': '100%','width': '100%', 'borderWidth': '1px', 'borderRadius': '5px', 'borderStyle': 'dashed'},
            persistence_type='session', persistence=True
        )
        return html.Div(
            id=self.id_,
            children=[timer, self.console, scroller],
            style={'height': '100%'}
        )

    def _register_callbacks(self, app):
        app.callback(
            [Output(self.console_id, 'value'),
             Output(self.js_id, 'run')],
            [Input(self.timer_id, 'n_intervals')],
            [State(self.console_id, 'value')]
        )(self.update_console)

    def monitor(self, process):
        self.process = process
        # Because there are only blocking reads to the pipe,
        # we need to read them on a separate thread.
        self._queue = queue.Queue()
        self._thread = threading.Thread(target=enqueue_output, args=(self.process.stdout, self._queue), daemon=True)
        self._thread.start()

    def call(self, command: str):
        command = shlex.split(command)
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.monitor(process)

    def update_console(self, _, current_text):
        if self.process is None:
            return [None, None]

        # We want to update the text field if there is new output from the process,
        # or if we detect the text value has been reset (due to e.g. switching tabs).
        try:
            line = self._queue.get_nowait()
            self._lines.append(line.decode('utf-8'))
        except queue.Empty:
            # No new message, update only required if value field had been reset.
            if current_text is not None:
                raise PreventUpdate

        return [''.join(self._lines), self.autoscroll_script if self.auto_scroll else '']
